@@ -8,6 +8,7 @@
  * Development of this code funded by Astaro AG (http://www.astaro.com/)
  */
 
+#define _GNU_SOURCE
 #include <config.h>
 #include <stdio.h>
 #include <string.h>
@@ -27,7 +28,7 @@ const struct location internal_location = {
 	.indesc	= &internal_indesc,
 };
 
-static const char *error_record_names[] = {
+static const char * const error_record_names[] = {
 	[EREC_INFORMATIONAL]	= NULL,
 	[EREC_WARNING]		= "Warning",
 	[EREC_ERROR]		= "Error"
@@ -36,10 +37,13 @@ static const char *error_record_names[] = {
 void erec_add_location(struct error_record *erec, const struct location *loc)
 {
 	assert(erec->num_locations < EREC_LOCATIONS_MAX);
-	erec->locations[erec->num_locations++] = *loc;
+	erec->locations[erec->num_locations] = *loc;
+	erec->locations[erec->num_locations].indesc = loc->indesc ?
+						    : &internal_indesc;
+	erec->num_locations++;
 }
 
-static void erec_destroy(struct error_record *erec)
+void erec_destroy(struct error_record *erec)
 {
 	xfree(erec->msg);
 	xfree(erec);
@@ -77,16 +81,64 @@ struct error_record *erec_create(enum error_record_types type,
 	return erec;
 }
 
-void erec_print(FILE *f, const struct error_record *erec)
+void print_location(FILE *f, const struct input_descriptor *indesc,
+		    const struct location *loc)
 {
-	const struct location *loc = erec->locations, *iloc;
-	const struct input_descriptor *indesc = loc->indesc, *tmp;
-	const char *line = NULL; /* silence gcc */
-	char buf[1024];
+	const struct input_descriptor *tmp;
+	const struct location *iloc;
+
+	if (indesc->location.indesc != NULL) {
+		const char *prefix = "In file included from";
+		iloc = &indesc->location;
+		for (tmp = iloc->indesc;
+		     tmp != NULL && tmp->type != INDESC_INTERNAL;
+		     tmp = iloc->indesc) {
+			fprintf(f, "%s %s:%u:%u-%u:\n", prefix,
+				tmp->name,
+				iloc->first_line, iloc->first_column,
+				iloc->last_column);
+			prefix = "                 from";
+			iloc = &tmp->location;
+		}
+	}
+	if (indesc->type != INDESC_BUFFER && indesc->name) {
+		fprintf(f, "%s:%u:%u-%u: ", indesc->name,
+			loc->first_line, loc->first_column,
+			loc->last_column);
+	}
+}
+
+const char *line_location(const struct input_descriptor *indesc,
+			  const struct location *loc, char *buf, size_t bufsiz)
+{
+	const char *line = NULL;
+	FILE *f;
+
+	f = fopen(indesc->name, "r");
+	if (!f)
+		return NULL;
+
+	if (!fseek(f, loc->line_offset, SEEK_SET) &&
+	    fread(buf, 1, bufsiz - 1, f) > 0) {
+		*strchrnul(buf, '\n') = '\0';
+		line = buf;
+	}
+	fclose(f);
+
+	return line;
+}
+
+void erec_print(struct output_ctx *octx, const struct error_record *erec,
+		unsigned int debug_mask)
+{
+	const struct location *loc = erec->locations;
+	const struct input_descriptor *indesc = loc->indesc;
+	const char *line = NULL;
+	char buf[1024] = {};
 	char *pbuf = NULL;
 	unsigned int i, end;
-	int l, ret;
-	off_t orig_offset = 0;
+	FILE *f;
+	int l;
 
 	switch (indesc->type) {
 	case INDESC_BUFFER:
@@ -94,15 +146,13 @@ void erec_print(FILE *f, const struct error_record *erec)
 		line = indesc->data;
 		*strchrnul(line, '\n') = '\0';
 		break;
+	case INDESC_STDIN:
+		line = indesc->data;
+		line += loc->line_offset;
+		*strchrnul(line, '\n') = '\0';
+		break;
 	case INDESC_FILE:
-		memset(buf, 0, sizeof(buf));
-		orig_offset = lseek(indesc->fd, 0, SEEK_CUR);
-		lseek(indesc->fd, loc->line_offset, SEEK_SET);
-		ret = read(indesc->fd, buf, sizeof(buf) - 1);
-		if (ret > 0)
-			*strchrnul(buf, '\n') = '\0';
-		line = buf;
-		lseek(indesc->fd, orig_offset, SEEK_SET);
+		line = line_location(indesc, loc, buf, sizeof(buf));
 		break;
 	case INDESC_INTERNAL:
 	case INDESC_NETLINK:
@@ -111,6 +161,8 @@ void erec_print(FILE *f, const struct error_record *erec)
 		BUG("invalid input descriptor type %u\n", indesc->type);
 	}
 
+	f = octx->error_fp;
+
 	if (indesc->type == INDESC_NETLINK) {
 		fprintf(f, "%s: ", indesc->name);
 		if (error_record_names[erec->type])
@@ -118,34 +170,21 @@ void erec_print(FILE *f, const struct error_record *erec)
 		fprintf(f, "%s\n", erec->msg);
 		for (l = 0; l < (int)erec->num_locations; l++) {
 			loc = &erec->locations[l];
-			netlink_dump_expr(loc->nle);
+			if (!loc->nle)
+				continue;
+			netlink_dump_expr(loc->nle, f, debug_mask);
 		}
-		fprintf(f, "\n");
-	} else {
-		if (indesc->location.indesc != NULL) {
-			const char *prefix = "In file included from";
-			iloc = &indesc->location;
-			for (tmp = iloc->indesc;
-			     tmp != NULL && tmp->type != INDESC_INTERNAL;
-			     tmp = iloc->indesc) {
-				fprintf(f, "%s %s:%u:%u-%u:\n", prefix,
-					tmp->name,
-					iloc->first_line, iloc->first_column,
-					iloc->last_column);
-				prefix = "                 from";
-				iloc = &tmp->location;
-			}
-		}
-		if (indesc->name != NULL)
-			fprintf(f, "%s:%u:%u-%u: ", indesc->name,
-				loc->first_line, loc->first_column,
-				loc->last_column);
-		if (error_record_names[erec->type])
-			fprintf(f, "%s: ", error_record_names[erec->type]);
-		fprintf(f, "%s\n", erec->msg);
+		return;
+	}
 
-		if (indesc->type != INDESC_INTERNAL)
-			fprintf(f, "%s\n", line);
+	print_location(f, indesc, loc);
+
+	if (error_record_names[erec->type])
+		fprintf(f, "%s: ", error_record_names[erec->type]);
+	fprintf(f, "%s\n", erec->msg);
+
+	if (line) {
+		fprintf(f, "%s\n", line);
 
 		end = 0;
 		for (l = erec->num_locations - 1; l >= 0; l--) {
@@ -154,6 +193,10 @@ void erec_print(FILE *f, const struct error_record *erec)
 		}
 		pbuf = xmalloc(end + 1);
 		memset(pbuf, ' ', end + 1);
+		for (i = 0; i < end && line[i]; i++) {
+			if (line[i] == '\t')
+				pbuf[i] = '\t';
+		}
 		for (l = erec->num_locations - 1; l >= 0; l--) {
 			loc = &erec->locations[l];
 			for (i = loc->first_column ? loc->first_column - 1 : 0;
@@ -167,13 +210,14 @@ void erec_print(FILE *f, const struct error_record *erec)
 	fprintf(f, "\n");
 }
 
-void erec_print_list(FILE *f, struct list_head *list)
+void erec_print_list(struct output_ctx *octx, struct list_head *list,
+		     unsigned int debug_mask)
 {
 	struct error_record *erec, *next;
 
 	list_for_each_entry_safe(erec, next, list, list) {
 		list_del(&erec->list);
-		erec_print(f, erec);
+		erec_print(octx, erec, debug_mask);
 		erec_destroy(erec);
 	}
 }
